#!/usr/bin/env python
#Once upon a time...

#Create a bam file with all the putatative telomere reads

import sys,os
import argparse
import textwrap
import pysam
import time
import gc
import pdb

import parabam

import Queue as Queue2

from multiprocessing import Queue,Process
from itertools import izip

######################################################################
##
##      Create telbams for list of given BAM files.
##
##      Author: jhrf
##
######################################################################

#Here we define what constitutes a telomereic read.
#If both reads in a pair match the critea we store
#them to the telbam
def rule(reads, constants, master):
    tel_pats = constants["tel_pats"]
    telomere_status = False

    for read in iter(reads):
        for pattern in tel_pats:
            if pattern in read.seq:
                telomere_status = True
                break

    results = []
    if telomere_status:
        results = [("telbam",reads[0]),("telbam",reads[1])]
    return results

#This whole package is essentially just a wrapper for a call to parabam subset
class Bam2Telbam(parabam.core.Interface):
    '''Interface to interact with telomerecat programatically. This interface
    is also called by the `telomerecat` script'''
    def __init__(self,
                 temp_dir=None,
                 task_size=250000,
                 total_procs=8,
                 reader_n=2,
                 verbose=False,
                 announce=True,
                 cmd_run=False):

        super(Bam2Telbam,self).__init__(instance_name = \
                                            "telomerecat bam2telbam", 
                                        temp_dir=temp_dir,
                                        task_size=task_size,
                                        total_procs=total_procs,
                                        reader_n=reader_n,
                                        verbose=verbose,
                                        announce = announce,
                                        cmd_run=cmd_run)

    def run_cmd(self):
        """Called from the master `telomerecat` script which handels the
        cmd line interface. Requires an argparse parser. Users should call
        the `run` function."""
        self.run(input_paths = self.cmd_args.input)

    def run(self,input_paths, keep_in_temp = False):
        """The main function for invoking the part of the 
           program which creates a telbam from a bam

        Arguments:
            bams (list): The BAM files we wish to run telomerecat telbam on
            total_procs (int): The maximum numbers of task that will be run at one time
            task_size (int): The amount of reads that any one task will process concurrently
            verbose (int): Expects an int from 0 to 2. 
                           The level of output produced by telomerecat
            keep_in_temp (bool): Files will be kept in temp file after processing. 
                                 Useful for incorporation into pipelines
            announce (bool): Specify whether the program returns a welcome string."""

        self.__introduce__()

        subset_types=["telbam"]
        tel_pats = ["TTAGGGTTAGGG","CCCTAACCCTAA"]

        #need to define my constants and engine here:
        telbam_constants = {"thresh":1,
                            "tel_pats":tel_pats}

        final_output_paths = {}

        for input_path in input_paths:

            if self.verbose:
                sys.stdout.write(" Generating TELBAM from: %s\n" % (input_path,))
                sys.stdout.write("\t- TELBAM generation started %s\n" %\
                                    (self.__get_date_time__(),))

            subset_interface = parabam.Subset(temp_dir=self.temp_dir,
                                              total_procs=self.total_procs,
                                              task_size=self.task_size,
                                              reader_n=self.reader_n,
                                              verbose=self.verbose,
                                              pair_process=True,
                                              include_duplicates=True,
                                              keep_in_temp=keep_in_temp)
            #call to parabam subset
            telbam_paths = subset_interface.run(input_paths=[input_path],
                                                subsets=subset_types,
                                                constants = telbam_constants,
                                                rule = rule)
            if self.verbose:
                sys.stdout.write("\t- TELBAM generation finished %s\n\n"\
                                     % (self.__get_date_time__(),))

            gc.collect()
            final_output_paths.update(telbam_paths)

        self.__goodbye__()
        return final_output_paths

    def get_parser(self):
        parser = self.default_parser()
        parser.description = textwrap.dedent('''\
        %s
        %s

             The bam2telbam command allows you to generate a TELBAM 
             from a parent BAM file. A TELBAM is a file including all of the
             reads with at least 2 occurences of the telomeric hexamer.

             Once you have genereated a TELBAM you may then generate length
             estimates more quickly, when compared to running the `bam2length`
             command on a full BAM file. This is helpful if you intend to 
             generate TL estimates more than once or if you require a  
             collection of telomere reads.
   
             Given the following BAM file:

                 example_bam_name.bam

             `telomerecat bam2telbam` will create the following TELBAM in the
             directory which it is run:

                 example_bam_name_telbam.bam

             To find out how to generate a length estimate from a TELBAM
             type `telomerecat telbam2length` into your terminal

        %s
         ''' % (self.instance_name, 
               self.header_line, 
               self.header_line,))

        parser.add_argument('input',metavar='BAM(S)', nargs='+',
            help='BAM file(s) for which we wish to generate TELBAMS')

        return parser


if __name__ == "__main__":
    print "Type telomerecat -h for help!"

#....happily ever after.



#!/usr/bin/env python
#Once upon a time...

import textwrap
import parabam

from telomerecat.core import TelomerecatInterface

######################################################################
##
##      Genereate length estimates for given BAM files
##
##      This simply strings together bam2telbam 
##      and telbam2length packages
##        
##      Author: jhrf
##
######################################################################

class Bam2Length(TelomerecatInterface):
    def __init__(self,temp_dir=None,
                 task_size=250000,
                 total_procs=8,
                 reader_n=1,
                 verbose=False,
                 announce=True,
                 cmd_run=False):

        super(Bam2Length,self).__init__(instance_name="telomerecat bam2length",
                                         temp_dir=temp_dir,
                                         task_size=task_size,
                                         total_procs=total_procs,
                                         reader_n=reader_n,
                                         verbose=verbose,
                                         announce=announce,
                                         cmd_run=cmd_run)

    def run_cmd(self):
        self.run(input_paths = self.cmd_args.input,
                 output_path = self.cmd_args.output,
                 inserts_path = self.cmd_args.insert,
                 discard_telbams = self.cmd_args.discard_telbams)

    def run(self, input_paths, 
                  output_path=None, 
                  inserts_path=None,
                  discard_telbams=False):
        
        #Import here to avoid infinite loop on import
        from telomerecat import Bam2Telbam
        from telomerecat import Telbam2Length

        self.__introduce__()

        telbam_interface = Bam2Telbam(temp_dir=self.temp_dir,
                                      total_procs=self.total_procs,
                                      task_size=self.task_size,
                                      reader_n=self.reader_n,
                                      verbose=self.verbose,
                                      announce=False)

        out_files = telbam_interface.run(input_paths=input_paths, 
                                         keep_in_temp = discard_telbams)

        length_paths = self.collapse_out_files(out_files)
        length_interface = Telbam2Length(temp_dir=self.temp_dir,
                                         total_procs=self.total_procs,
                                         task_size=self.task_size,
                                         reader_n = self.reader_n,
                                         announce=False,
                                         verbose=self.verbose)

        length_interface.run(input_paths=length_paths,
                             output_path=output_path,
                             inserts_path=inserts_path)

        self.__goodbye__()
        self.interface_exit()

    def collapse_out_files(self,out_files):
        length_paths = []
        for path in out_files.keys():
            length_paths.append(out_files[path]["telbam"])
        return length_paths

    def get_parser(self):
        parser = self.default_parser()
        parser.description = textwrap.dedent(
        '''\
        %s
        %s

            The bam2length command allows the user to genereate a telomere
            length estimate from a BAM file.

            The majority of the time taken running the bam2length script is
            spent collecting telomeric reads from the BAM file. If you wish
            to run multiple analyses on this file be sure to keep the TELBAMS
            that this run creates by using the `-k` option.

            If you wish to generate TELBAMS seperately from length estimation
            you should use the bam2telbam command.
            
            Type `telomerecat bam2telbam` to find out more.

        %s
        ''' % (self.instance_name, 
               self.header_line, 
               self.header_line,))        

        parser.add_argument('input',metavar='BAM(S)', nargs='+',
            help=('BAM file(s) for which we wish to\n'
                  'generate telomere length estimates'))
        parser.add_argument('-x','--discard_telbams',
            action="store_true",default=False,
            help=('The program will NOT save any TELBAMs\n'
                   'generated as part of the analysis'))

        return parser

if __name__ == "__main__":
    print "Please do not run this script directly."\
     " Type telomerecat -h for more information."


#....happily ever after.


#__init__.py
import core

from telomerecat.csv2length import Csv2Length
from telomerecat.bam2telbam import Bam2Telbam
from telomerecat.telbam2length import Telbam2Length
from telomerecat.bam2length import Bam2Length

import bam2telbam
import bam2length
import telbam2length
import csv2length

from _version import __version__


import sys

import parabam
from abc import ABCMeta, abstractmethod

class TelomerecatInterface(parabam.core.Interface):
    __metaclass__ = ABCMeta

    def __init__(self,
                 instance_name,
                 temp_dir=None,
                 task_size=10000,
                 total_procs=8,
                 reader_n=2,
                 verbose=False,
                 announce=True,
                 cmd_run=False):

        super(TelomerecatInterface,self).__init__(
                                instance_name = instance_name,
                                temp_dir=temp_dir,
                                task_size=task_size,
                                total_procs=total_procs,
                                reader_n=reader_n,
                                verbose=verbose,
                                announce = announce,
                                cmd_run=cmd_run)

        
    
    def __output__(self,outstr,level=-1):
        if self.verbose and (self.verbose >= level or level == -1):
            sys.stdout.write(outstr)
            sys.stdout.flush()

    def default_parser(self):
        parser = super(TelomerecatInterface,self).default_parser()

        parser.add_argument('--output',
            metavar='CSV',type=str,nargs='?',default=None,
            help=('Specify output path for length estimation CSV.\n'
                  'Path automatically generated if no path is\n'
                  'specified [Default: None]'))
        parser.add_argument('--insert',
            metavar='CSV',nargs='?',type=str,default=None,
            help=('A file specifying the insert length mean and\n'
                 'std for each input sample. If not present\n'
                 'telomerecat will automatically estimate\n'
                 'insert length of sample [Default: None]'))
        parser.add_argument('-w','--prior_weight',
            metavar='INT',type=int,default=3,
            help=('The weight given to the prior expectation\n'
                  'in F2a correction [Default 3]'))
        parser.add_argument('-N','--simulator_runs',
            metavar='INT',type=int,default=10,
            help=('The amount of times to run the length simulator.\n'
                  'A higher number better captures the uncertainty \n'
                  'produced by the insert length\n'
                  'distribution [Deafult 10]'))
        parser.add_argument('-d','--disable_correction',
            action="store_true",default=False
            ,help='Correction will not be applied to F2a values')
        parser.add_argument('-b','--brief_csv',
            action="store_true",default=False
            ,help=('The output CSV will be comprised of\n'
                   'only the essential columns'))

        return parser




import sys
import textwrap
import time
import os
import re
import parabam
import pdb

import numpy as np
import pandas as pd

from shutil import copy
from itertools import izip
from collections import namedtuple,Counter
from pprint import pprint

from telomerecat import Csv2Length
from telomerecat.core import TelomerecatInterface

######################################################################
##
##      Create a length estimate given a set of TELBAMS 
##
##      Author: jhrf
##
######################################################################

class SimpleReadFactory(object):

    def __init__(self, vital_stats=None, trim_reads=0):
        self._SimpleRead = namedtuple("SimpleRead","seq qual" +
                                      " five_prime pattern mima_loci n_loci"+
                                      " avg_qual")

        if vital_stats:
            self._read_len = vital_stats["read_len"]
            self._phred_offset = vital_stats["phred_offset"]
        else:
            self._read_len = 100
            self._phred_offset = 33

        self._trim_reads = trim_reads
        self._compliments = {"A":"T","T":"A",
                             "C":"G","G":"C",
                             "N":"N"}

        self.mima_logic = MismatchingLociLogic()

    def get_simple_read(self,read):
        seq,qual = self.__flip_and_compliment__(read)
        if self._trim_reads > 0:
            seq,qual = (seq[:self._trim_reads],
                        qual[:self._trim_reads])

        (mima_loci,frameshift_loci),pattern = \
                self.mima_logic.get_telo_mismatch(seq)

        avg_qual,n_loci = self.__get_phred_score__(qual, 
                                                   mima_loci, 
                                                   frameshift_loci)

        simple_read = self._SimpleRead(
            seq,
            qual,
            self.__get_five_prime__(pattern),
            pattern,
            mima_loci,
            n_loci,
            avg_qual)

        return simple_read

    def __get_phred_score__(self, qual, mima_loci, frameshift_loci):
        if len(mima_loci) + len(frameshift_loci) == 0:
            return 0, 0

        remove_mimas = []
        phreds = []

        for start, end in frameshift_loci:
            fuse_phreds = []

            for i in xrange(start,end):
                if i in mima_loci:
                    remove_mimas.append(i)
                fuse_phreds.append(ord(qual[i])-self._phred_offset)

            phreds.append(min(fuse_phreds))

        for loci in mima_loci:
            if loci not in remove_mimas:
                phreds.append(ord(qual[loci])-self._phred_offset)

        return np.mean(phreds), len(phreds)

    def __get_average_qual__(self,qual,mima_loci):
        if len(mima_loci) == 0:
            return 0
        phreds = np.array([ord(qual[i])-self._phred_offset for i in mima_loci])
        return np.mean(phreds)

    def __trim_seq__(self,seq,qual):
        cutoff = 0
        min_sequence = 0
        for q in qual:
            qual_byte = ord(q) - self._phred_offset
            if qual_byte == 0:
                min_sequence += 1
            else:
                min_sequence = 0
            
            if min_sequence == 5:
                cutoff = cutoff - 5
                break

            cutoff += 1

        return seq[:cutoff],qual[:cutoff]

    def __get_five_prime__(self,pattern):
        if pattern is None:
            return None
        else:
            return pattern == "CCCTAA"

    def __get_pattern__(self,seq):
        cta,tag = "CCCTAA","TTAGGG"
        pattern = None
        if cta in seq or tag in seq:   
            if seq.count(cta) > seq.count(tag):
                pattern = cta
            else:
                pattern = tag
        return pattern

    def __flip_and_compliment__(self,read):
        if read.is_reverse:
            compliments = self._compliments
            seq_compliment = map(lambda base: compliments[base],read.seq)
            seq_compliment = "".join(seq_compliment)
            return(seq_compliment[::-1],read.qual[::-1])
        else:
            return (read.seq,read.qual)

class MismatchingLociLogic(object):

    def get_telo_mismatch(self, seq):

        c_rich_count = seq.count("CCCTAA")
        g_rich_count = seq.count("TTAGGG")

        if c_rich_count > g_rich_count:
            return self.get_mismatching_loci(seq, "CCCTAA"), "CCCTAA"
        elif g_rich_count > c_rich_count:
            return self.get_mismatching_loci(seq, "TTAGGG"), "TTAGGG"
        else:
            c_mima, c_fuse = self.get_mismatching_loci(seq, "CCCTAA")
            g_mima, g_fuse = self.get_mismatching_loci(seq, "TTAGGG")

            c_score = len(c_mima) + len(c_fuse)
            g_score = len(g_mima) + len(g_fuse)

            if c_score < g_score:
                return (c_mima, c_fuse), "CCCTAA"
            else:
                return (g_mima, g_fuse), "TTAGGG"

    def get_mismatching_loci(self, seq, pattern):
        segments = re.split("(%s)"%(pattern,), seq)
        segments = self.filter_segments(segments)
        segments = self.collapse_segments(segments, pattern)
        segments = self.merge_segments(segments, pattern)
        mima_loci, fuse_loci = self.find_mima_loci(segments, pattern)

        return mima_loci, fuse_loci

    def find_mima_loci(self, segments, pattern):
        mima_loci = []
        fuse_loci = []

        offset = 0

        for i,seg in enumerate(segments):
            neighbours = self.get_neighbour_segments(segments,i)
            
            score, segment_loci = \
                self.get_best_offset_score(seg, pattern, neighbours)
            
            mima_loci.extend(np.array(segment_loci)+offset)

            if pattern in neighbours[0] or pattern in seg:
                if offset-1 not in mima_loci and \
                    offset+1 not in mima_loci and \
                     offset not in mima_loci:

                    if neighbours[0] != "":
                        fuse_loci.append( (offset-1,offset+2,))
            offset += len(seg)

        return mima_loci,fuse_loci

    def get_neighbour_segments(self, segments, i):
        prev_segment = ""
        next_segment = ""

        if i > 0:
            prev_segment = segments[i-1]
        if i < (len(segments)-1):
            next_segment = segments[i+1]

        return prev_segment, next_segment

    def get_best_offset_score(self, seq, 
                                    pattern, 
                                    neighbours, 
                                    override = False):

        best_score = float("Inf")
        best_mima = []

        if pattern in seq and not override:
            best_score, best_mima = 0,[]
        elif len(seq) == 1:
            best_score, best_mima =  1,[0]
        elif len(seq) < len(pattern)*.66:
            for position,neighbour in enumerate(neighbours):
                if len(neighbour) > 0:
                    if position == 0:
                        temp_seq = neighbour+seq
                    else:
                        temp_seq = seq+neighbour
                    score, mima = self.get_best_offset_score(temp_seq, 
                                                             pattern,
                                                             neighbours,
                                                             override=True)
                    if score < best_score:
                        best_score = score
                        if position == 0:
                            best_mima = \
                                (np.array(mima) - len(neighbour)).tolist()
                        else:
                            best_mima = mima

        else:
            for i in xrange(len(pattern)):
                comparison_seq = self.telo_sequence_generator(pattern, i)
                mima_loci = []
                score = 0
                for s, (l, c) in izip(seq, comparison_seq):
                    if s != c:
                        mima_loci.append(l)
                        score += 1
                        if score > best_score:
                            break

                if score < best_score:
                    best_score = score
                    best_mima = mima_loci

        return best_score, best_mima
            
    def telo_sequence_generator(self, pattern, offset):

        offset_pattern = (pattern[offset:] + pattern)[:len(pattern)]
        i = 0
        while True:
            yield i, offset_pattern[i % len(pattern)]
            i += 1

    def filter_segments(self, segments):
        return [seg for seg in segments if seg != '']

    def merge_segments(self, segments, pattern):
        merge_segments = list(segments)

        for i,seg in enumerate(segments):
            if pattern not in seg:
                top_trim = 0
                bottom_trim = len(seg)
                
                #See if valid sequence continues into the corrupted segment
                if i > 0:
                    #Check preccedding segment
                    top_trim = self.compare_to_pattern(seg, pattern)
                if i < len(segments)-1:
                    #Check succeeding segment
                    bottom_trim = self.compare_to_pattern(seg,
                                                          pattern,
                                                          reverse = True)

                #ensure that loci are not merged into both
                # i-1 and i+1 simealteanously
                if bottom_trim < top_trim:
                    bottom_trim = top_trim

                #do merge operation
                if top_trim > 0:
                    merge_segments[i-1] += seg[:top_trim]
                if bottom_trim < len(seg):
                    merge_segments[i+1] = seg[bottom_trim:] +\
                                             merge_segments[i+1]
            
                #correct current segment
                merge_segments[i] = seg[top_trim:bottom_trim]

        return self.filter_segments(merge_segments)

    def compare_to_pattern(self, seq, pattern, reverse = False):
        i = 0
        generator = self.get_sequence_generator(seq, pattern, reverse)

        for c,s in generator:
            if c == s:
                i += 1
            else:
                break

        if reverse:
            return_index =  len(seq) - i
        else:
            return_index = i

        return return_index

    def get_sequence_generator(self, seq, pattern, reverse): 
        def forward_gen(seq, pattern):
            for c,s in zip(seq,pattern):
                yield c, s
            return

        def reverse_gen(seq, pattern):
            for c,s in izip(seq[::-1], pattern[::-1]):
                yield c, s
            return

        if reverse:
            generator = reverse_gen(seq, pattern) 
        else: 
            generator = forward_gen(seq, pattern) 

        return generator

    def collapse_segments(self, segments, pattern):
        collapsed_segments = []
        current_segment = ''
        for seg in segments:
            if seg == pattern:
                current_segment += seg
            else:
                if current_segment != '':
                    collapsed_segments.append(current_segment)
                collapsed_segments.append(seg)
                current_segment = ''

        if current_segment != '':
            collapsed_segments.append(current_segment)

        return collapsed_segments

    def print_mima(self, seq, qual, pat):
        print "-"
        loci_status, mima_loci, fuse_loci =\
            self.get_loci_status(seq, pat)

        print "Mima:",mima_loci
        print "Fuse:",fuse_loci
        print seq
        print loci_status
        print qual

    def get_loci_status(self, seq, pat):
        mima_loci,fuse_loci = self.get_mismatching_loci(seq,pat)
        loci_status = []

        for i in xrange(len(seq)):
            if i in mima_loci:
                loci_status.append("X")
            else:
                loci_status.append("_")

        for start,end in fuse_loci:
            loci_status[start:end] = ["F"] * (end - start)

        return "".join(loci_status), mima_loci, fuse_loci

class VitalStatsFinder(object):

    def __init__(self,temp_dir,total_procs,task_size,trim_length=0):
        self.temp_dir = temp_dir
        self._total_procs = total_procs
        self._task_size = task_size
        self._trim_length = trim_length

    def __csv_to_dict__(self,stats_path):
        insert_dat = np.genfromtxt(stats_path,delimiter=",",
                                names=True,dtype=("S256",float,float,
                                                         float,float,
                                                         float,float,
                                                         float,float,
                                                         float))

        ins_N = int(insert_dat['N'])
        if ins_N == 0:
            insert_mean = -1
            insert_sd = -1
        else:
            ins_sum = int(insert_dat['sum'])
            ins_power_2 = int(insert_dat['power_2'])

            insert_mean,insert_sd = \
                        self.__get_mean_and_sd__(ins_sum, ins_power_2, ins_N)
    
        min_qual = int(insert_dat['min_qual'])
        qual_mean,qual_sd = self.__get_mean_and_sd__(insert_dat["qual_sum"],
                                                    insert_dat["qual_power_2"],
                                                    insert_dat["qual_N"])

        return {"insert_mean":insert_mean, 
                "insert_sd": insert_sd,
                "min_qual":min_qual,
                "max_qual":int(insert_dat['max_qual']),
                "read_len":int(insert_dat['read_len']),
                "qual_mean":qual_mean,
                "qual_sd":qual_sd}

    def __get_mean_and_sd__(self,x_sum,x_power_2,x_N):
        x_mean = x_sum / x_N
        x_sd = np.sqrt( (x_N * x_power_2) - x_sum**2) / x_N

        return x_mean,x_sd

    def get_vital_stats(self,sample_path):

        vital_stats_csv = self.__run_vital_rule__(sample_path)
        vital_stats = self.__csv_to_dict__(vital_stats_csv)
        vital_stats["phred_offset"] = vital_stats["min_qual"]
        vital_stats["initial_read_len"] = vital_stats["read_len"]

        if self._trim_length > 0:
            vital_stats["read_len"] = self._trim_length 
 
        return vital_stats

    def __run_vital_rule__(self,sample_path,keep_in_temp=True):
        def rule(read,constants,master):
            stats = {}

            hash_count = read.qual.count("#")
            if read.is_read1 and read.is_proper_pair and read.mapq > 38:
                insert_size = abs(read.template_length)
                stats["sum"] = insert_size
                stats["power_2"] = insert_size**2
                stats["N"] = 1
            
            stats["read_len"] =  len(read.seq)
            byte_vals = map(ord,read.qual)
            min_qual = min(byte_vals)
            max_qual = max(byte_vals)

            qual_mean = np.mean(byte_vals)
            stats["qual_sum"] = qual_mean
            stats["qual_power_2"] = qual_mean**2
            stats["qual_N"] = 1

            stats["min_qual"] = min_qual
            stats["max_qual"] = max_qual

            return stats

        structures = {}

        structures["sum"] = {"data":0,"store_method":"cumu"}
        structures["power_2"] = {"data":0,"store_method":"cumu"}
        structures["N"] = {"data":0,"store_method":"cumu"}
        structures["read_len"] = {"data":0,"store_method":"max"}

        structures["min_qual"] = {"data":999,"store_method":"min"}
        structures["max_qual"] = {"data":0,"store_method":"max"}


        structures["qual_sum"] = {"data":0,"store_method":"cumu"}
        structures["qual_power_2"] = {"data":0,"store_method":"cumu"}    
        structures["qual_N"] = {"data":0,"store_method":"cumu"}  

        stat_interface = parabam.Stat(temp_dir=self.temp_dir,
                                      total_procs = self._total_procs,
                                      task_size = 10000,
                                      keep_in_temp=keep_in_temp)

        out_paths = stat_interface.run(input_paths= [sample_path],
                                       constants = {},
                                       rule = rule,
                                       struc_blueprint = structures)

        return out_paths["global"]["stats"]

class ReadStatsFactory(object):

    def __init__(self,temp_dir,
                      total_procs=4,
                      task_size = 5000,
                      trim_reads = 0,
                      debug_print=False):

        self.temp_dir = temp_dir
        self._total_procs = total_procs
        self._task_size = task_size

        self._debug_print = debug_print
        self._trim = trim_reads

    def get_read_counts(self,path,vital_stats):
        read_stat_paths = self.run_read_stat_rule(path, vital_stats)

        read_array = self.__path_to_read_array__(read_stat_paths["read_array"])

        error_profile, sample_variance = \
                    self.__paths_to_error_profile__(read_stat_paths)

        read_counts = self.read_array_to_counts(read_array,
                                                error_profile,
                                                sample_variance)

        self.__delete_analysis_paths__(read_stat_paths)
        
        return read_counts

    def __delete_analysis_paths__(self, read_stat_paths):
        for analysis, path in read_stat_paths.items():
            os.remove(path)

    def __paths_to_error_profile__(self,read_stat_paths):
        random_counts = pd.read_csv(read_stat_paths["random_counts"],
                                    header=None).values 
        read_counts = pd.read_csv(read_stat_paths["mima_counts"],
                                  header=None).values 
        error_profile = self.__array_to_profile__(read_counts, random_counts)
        sample_variance = self.__get_sample_variance__(read_counts)
        
        return error_profile, sample_variance

    def __get_sample_variance__(self, read_counts):
        read_counts[0,:] = 0
        mask = read_counts > 0
        mask[40:,:] = False

        read_variance = (read_counts[mask].std() / read_counts[mask].mean())
        return read_variance

    def __array_to_profile__(self,read_counts,random_counts, thresh = None):
        dif_counts = read_counts - random_counts
        ten_percent = int(read_counts.shape[0] * .1)

        if thresh is None:
            mask = self.__get_exclusion_mask__(read_counts)
            arg_max_index = (read_counts * mask).argmax()
            dif_loci_x,dif_loci_y = np.unravel_index(arg_max_index,
                                                     dif_counts.shape)

            hi_thresh = dif_counts[int(dif_loci_x-ten_percent):\
                                  int(dif_loci_x+ten_percent),
                                  dif_loci_y-15:\
                                  dif_loci_y+1]
            hi_thresh = hi_thresh.flatten()

            thresh = np.percentile(hi_thresh,95)

            print thresh

        if self._debug_print:
            print 'Thresh:',thresh

        error_profile = (dif_counts * (dif_counts > 0))\
                                 > thresh

        error_profile = self.__remove_noise__(error_profile)
        error_profile = self.__prune_error_profile__(error_profile)
        error_profile = self.__rationalise_error_profile__(error_profile)

        error_profile[:ten_percent,:] = 1

        return error_profile

    def __remove_noise__(self, error_profile):
        row_max, col_max = error_profile.shape
        error_profile[int(row_max*.11):,int(col_max*.7):] = 0
        error_profile[int(row_max*.35):,0] = 0
        error_profile[int(row_max*.55):,:] = 0 

        return error_profile

    def __get_exclusion_mask__(self, read_counts):
        mask = np.zeros(read_counts.shape)
        x_start = int(read_counts.shape[0] * .2)
        y_start = int(read_counts.shape[1] * .5)
        mask[x_start:,y_start:] = 1
        return mask

    def __prune_error_profile__(self, error_profile):
        first_locis = self.__get_first_loci__(error_profile)
        prune_mask = np.ones(error_profile.shape)
        for row_i in xrange(1,error_profile.shape[0]):
            if first_locis[row_i] == -1:
                continue 
            else:
                for col_i in xrange(error_profile.shape[1]):
                    if (not error_profile[row_i,col_i]) or \
                        col_i == 0:
                            continue
                    elif self.__prune_decision__(row_i,col_i,error_profile):
                        prune_mask[row_i,col_i] = 0
        return error_profile * prune_mask

    def __prune_decision__(self, row_i, col_i, error_profile):
        try:
            return self.__get_neighbor_sum__(row_i,col_i,error_profile) < 4
        except IndexError:
            return False

    def __get_neighbor_sum__(self,row_i, col_i, error_profile):

        neighbours = [(row_i-1,col_i+1),
                      (row_i-1,col_i),
                      (row_i-1,col_i-1),
                      (row_i,col_i+1),
                      (row_i,col_i-1),
                      (row_i+1,col_i+1),
                      (row_i+1,col_i),
                      (row_i+1,col_i-1),]

        neighbours_sum = sum([ error_profile[r,c] for (r,c) in neighbours])
        return neighbours_sum

    def __get_first_loci__(self,error_profile):
        first_loci = []
        for row_i in xrange(error_profile.shape[0]):
            if any(error_profile[row_i,:]):
                for col_i in xrange(error_profile.shape[1]):
                    if error_profile[row_i,col_i]:
                        first_loci.append(col_i)
                        break
            else:
                first_loci.append(-1)
        return first_loci

    def __rationalise_error_profile__(self, error_profile):
        start_row = np.where(error_profile)[0].max()
        global_loci = 0
        for i in reversed(xrange(0,start_row+1)):
            error_bins_in_row = np.where(error_profile[i,:])[0]
            if len(error_bins_in_row) > 0:
                cur_loci = error_bins_in_row.max()
            else:
                cur_loci = 0

            #if cur_loci > global_loci:
            global_loci = cur_loci
            error_profile[i,:global_loci+1] = True
        
        return error_profile

    def __array_to_file__(self,array,unique):
        df = pd.DataFrame(array)
        out_path = "./%s-tmctout.csv" % (unique)
        df.to_csv(out_path,index=False,header=False)
        return out_path

    def __path_to_read_array__(self,read_array_path):
        return pd.read_csv(read_array_path,header=None).values

    def read_array_to_counts(self, read_array, error_profile, sample_variance):
        complete_reads,boundary_reads = \
                        self.__get_complete_status__(read_array,error_profile)

        f2_count,f4_count = self.__get_boundary_counts__(boundary_reads)
        f1_count = self.__get_f1_count__(complete_reads)

        return_dat = { "F2":int(f2_count),
                       "F1":int(f1_count),
                       "F4":f4_count,
                       "sample_variance":sample_variance}

        return return_dat

    def __get_f1_count__(self,complete_reads):
        return float(complete_reads.shape[0]) / 2

    def __get_boundary_counts__(self,boundary_reads):
        f2_count,f4_count,total_reads = \
                 self.__get_read_counts__(boundary_reads)

        observed_f2a_ratio = (f2_count-f4_count) / float(total_reads)
        observed_f2a_weight = f2_count
            
        return f2_count, f4_count

    def __get_read_counts__(self,boundary_reads):
        f2_count = sum(boundary_reads[:,3] == 1)
        f4_count = sum(boundary_reads[:,3] == 0)
        total_reads = boundary_reads.shape[0]
        return f2_count,f4_count,total_reads
        
    def __get_complete_status__(self,read_array,error_profile):
        boundary_indicies = []
        complete_indicies = []

        for i in xrange(int(read_array.shape[0])):
            read_info = map(int,read_array[i,[0,-2]])
            pair_info = map(int,read_array[i,[2,-1]])

            read = error_profile[read_info[0],read_info[1]] 
            pair = error_profile[pair_info[0],pair_info[1]]

            if read and pair:
                complete_indicies.append(i)
            elif (not read) and pair:
                boundary_indicies.append(i)

        return read_array[complete_indicies,:],\
                read_array[boundary_indicies,:]

    def run_read_stat_rule(self,path,
                                vital_stats,
                                keep_in_temp=True):

        simple_read_factory = SimpleReadFactory(vital_stats,
                                                trim_reads=self._trim)
        phred_offset = vital_stats["phred_offset"]

        maxtrix_max = (vital_stats["max_qual"] - phred_offset)+1
        matrix_shape = (vital_stats["read_len"]+1,maxtrix_max)

        def get_return_stats(reads):

            return_stats = [len(reads[0].mima_loci),
                            int(reads[0].five_prime),
                            len(reads[1].mima_loci),
                            int(reads[1].five_prime),
                            reads[0].avg_qual,
                            reads[1].avg_qual]

            return return_stats

        def rule(reads, constants, master):
            simple_reads = [simple_read_factory.get_simple_read(read) \
                                                         for read in reads]
            return_dat = np.zeros((2,6))
            return_dat[0,:] = get_return_stats(simple_reads)
            return_dat[1,:] = get_return_stats(simple_reads[::-1])

            random_counts = np.zeros(matrix_shape)
            mima_counts = np.zeros(matrix_shape)

            for read in simple_reads:
                mima_counts[read.n_loci,int(read.avg_qual)] += 1

                #sample_size = int(np.random.uniform(1,80))
                sample_size = len(read.mima_loci)
                if sample_size > 0:
                    rand_quals  = np.random.choice(list(read.qual),sample_size)
                    qual_bytes  = [ord(q) - phred_offset for q in rand_quals]
                    rand_avg = np.mean(qual_bytes)

                    random_counts[int(sample_size),int(rand_avg)] += 1

            results = {"read_array":np.array(return_dat),
                      "random_counts":random_counts,
                      "mima_counts":mima_counts}

            return results

        structures = {"read_array":{"data":np.zeros((2,6)),
                                    "store_method":"vstack"},
                     "mima_counts":{"data":np.zeros(matrix_shape),
                                    "store_method":"cumu"},
                     "random_counts":{"data":np.zeros(matrix_shape),
                                    "store_method":"cumu"},}

        stat_interface = parabam.Stat(temp_dir=self.temp_dir,
                                      pair_process=True,
                                      total_procs = self._total_procs,
                                      task_size = self._task_size,
                                      keep_in_temp = keep_in_temp,
                                      verbose=0)

        out_paths = stat_interface.run(
            input_paths = [path],
            constants = {},
            rule = rule,
            struc_blueprint = structures)

        return out_paths[path]

class Telbam2Length(TelomerecatInterface):

    def __init__(self,
                 temp_dir=None,
                 task_size=10000,
                 total_procs=8,
                 reader_n=2,
                 verbose=False,
                 announce=True,
                 cmd_run=False):

        super(Telbam2Length,self).__init__(instance_name = \
                                            "telomerecat telbam2length", 
                                        temp_dir=temp_dir,
                                        task_size=task_size,
                                        total_procs=total_procs,
                                        reader_n=reader_n,
                                        verbose=verbose,
                                        announce = announce,
                                        cmd_run=cmd_run)

    def run_cmd(self):
        self.run(input_paths = self.cmd_args.input,
                 trim = self.cmd_args.trim,
                 inserts_path = self.cmd_args.insert,
                 correct_f2a = not self.cmd_args.disable_correction,
                 output_path = self.cmd_args.output)

    def run(self,input_paths,
                 trim = 0,
                 output_path = None,
                 correct_f2a = True,
                 inserts_path = None):
        
        """The main function for invoking the part of the 
           program which creates a telbam from a bam

        Arguments:
            inputs_paths (list): The TELBAMs that we wish to estimate TL estimates for
            output_path (string): Specify a path to output results to (optional)
            inserts_path (string): A path to a file containing insert length estimates
                                   for each TELBAM. Formatted as follows:
                                        example_telbam.bam, insert_mean, insert_sd
        """

        self.__introduce__()

        names = map(lambda b: os.path.basename(b),input_paths)
        names = map(lambda nm: nm.replace("_telbam",""),names)

        output_csv_path = self.__get_output_path__(output_path)
        temp_csv_path = self.__get_temp_path__()

        insert_length_generator = self.__get_insert_generator__(inserts_path)
        
        self.__output__(" Collecting meta-data for all samples | %s\n" \
                            % (self.__get_date_time__(),),1)

        vital_stats_finder = VitalStatsFinder(self.temp_dir, 
                                        self.total_procs,
                                        self.task_size,
                                        trim)

        for sample_path,sample_name, in izip(input_paths,names):
            sample_intro = "\t- %s | %s\n" % (sample_name, 
                                              self.__get_date_time__())

            self.__output__(sample_intro,2)

            vital_stats = vital_stats_finder.get_vital_stats(sample_path)

            self.__check_vital_stats_insert_size__(inserts_path,
                                                    insert_length_generator,
                                                    vital_stats)

            read_type_counts = self.__get_read_types__(sample_path,
                                                       vital_stats,
                                                       self.total_procs,
                                                       trim)

            self.__write_to_csv__(read_type_counts,
                                    vital_stats,
                                    temp_csv_path,
                                    sample_name)
        
        self.__output__("\n",1)
        length_interface = Csv2Length(temp_dir=self.temp_dir,
                                       total_procs=self.total_procs,
                                       verbose=self.verbose,
                                       announce=False,
                                       cmd_run=False)

        length_interface.run(input_paths=[temp_csv_path], 
                             output_paths=[output_csv_path],
                             correct_f2a=correct_f2a)

        self.__print_output_information__(output_csv_path)
        self.__goodbye__()

        return output_csv_path

    def __print_output_information__(self, output_csv_path):
        self.__output__((" Length estimation results "
                         "written to the following file:\n"),1)
        self.__output__("\t./%s\n\n" % (os.path.basename(output_csv_path,)))

    def __get_insert_generator__(self,inserts_path):
        if inserts_path:
            with open(inserts_path,"r") as inserts_file:
                for line in inserts_file:
                    yield map(float,line.split(","))

    def __check_vital_stats_insert_size__(self,inserts_path,
                                        insert_length_generator,vital_stats):
        if inserts_path:
            insert_mean,insert_sd = insert_length_generator.next()
            vital_stats["insert_mean"] = insert_mean
            vital_stats["insert_sd"] = insert_sd
            self.__output__("\t\t+ Using user defined insert size: %d,%d\n" \
                                                    % (insert_mean,insert_sd),2)
        elif vital_stats["insert_mean"] == -1:
            default_mean,default_sd = 350,25
            vital_stats["insert_mean"] = 350
            vital_stats["insert_sd"] = 25
            self.__output__("\t\t+ Failed to estimate insert size. Using default: %d,%d\n"\
                                                % (default_mean,default_sd),2)

    def __get_read_types__(self,sample_path,
                                vital_stats,
                                total_procs,
                                trim,
                                read_stats_factory=None):

        if read_stats_factory is None:
            read_stats_factory = ReadStatsFactory(temp_dir=self.temp_dir,
                                                  total_procs=total_procs,
                                                  trim_reads=trim,
                                                  debug_print=False)
            
        read_type_counts = read_stats_factory.get_read_counts(sample_path,
                                                              vital_stats)
        return read_type_counts 

    def __get_temp_path__(self):
        temp_path = os.path.join(self.temp_dir,"telomerecat_temp_%d.csv" \
                                                            % (time.time()))
        self.__create_output_file__(temp_path)
        return temp_path

    def __get_output_path__(self, user_output_path):
        tmct_output_path = None
        if user_output_path is None:
            tmct_output_path = os.path.join("./telomerecat_length_%d.csv" \
                                                            % (time.time(),))
        else:
            tmct_output_path = user_output_path

        return tmct_output_path

    def __create_output_file__(self,output_csv_path):
        with open(output_csv_path,"w") as total:
            header = ("Sample,F1,F2,F4,Psi,Insert_mean,Insert_sd,"
                      "Read_length,Initial_read_length\n")
            total.write(header)
        return output_csv_path

    def __write_to_csv__(self,
                         read_type_counts,
                         vital_stats,
                         output_csv_path,
                         name):
        with open(output_csv_path,"a") as counts:
            counts.write("%s,%d,%d,%d,%.3f,%.3f,%.3f,%d,%d\n" %\
                (name,
                read_type_counts["F1"],
                read_type_counts["F2"],
                read_type_counts["F4"],
                read_type_counts["sample_variance"],
                vital_stats["insert_mean"],
                vital_stats["insert_sd"],
                vital_stats["read_len"],
                vital_stats["initial_read_len"]))

    def get_parser(self):
        parser = self.default_parser()
        parser.description = textwrap.dedent(
        '''\
        %s
        %s

            The telbam2length command allows the user to genereate a telomere
            length estimate from a previously generated TELBAM file.

            Example useage:

            telomerecat telbam2length /path/to/some_telbam.bam

            This will generate a .csv file with an telomere length estimate
            for the `some_telbam.bam` file.

        %s
        ''' % (self.instance_name,
               self.header_line,
               self.header_line,))

        parser.add_argument('input', metavar='TELBAM(S)', nargs='+',
            help="The TELBAM(s) that we wish to analyse")
        parser.add_argument('--output',
            metavar='CSV',type=str, nargs='?', default=None,
            help=('Specify output path for length estimation CSV.\n'
                  'Automatically generated if left blank [Default: None]'))
        parser.add_argument('-s',type=int, nargs='?', default=10000
            ,help="The amount of reads considered by each\n"
                    "distributed task. [Default: 10000]")
        parser.add_argument('-t',"--trim", type=int, nargs='?', default=0
            ,help="Use only the amount of sequence specified by this  \n"
                  "option (i.e if the value 90 is supplied\n"
                  "then only the first 90 bases are\n"
                  "considered) [Default: Whole read]")

        return parser

if __name__ == "__main__":
    print "Do not run this script directly. Type `telomerecat` for help."



from setuptools import setup,find_packages
import os 
import sys
import shutil

#Instalise the __version__ variable
execfile('./telomerecat/_version.py')

setup(name='telomerecat',
	description='Telomere Computational Analysis Tool',
      version=__version__,
      author="JHR Farmery",
      license='GPL',
      author_email = 'jhrf2@cam.ac.uk',
      packages = ['telomerecat'],
      package_dir = {'telomerecat':'telomerecat'},
      install_requires = ['parabam>=2.2','argparse','numpy','pysam','pandas'],
      include_package_data = True,
      scripts = ['./telomerecat/bin/telomerecat'],
      zip_safe = False
    )



import sys
import textwrap
import time
import os
import parabam
import random
import pdb

import numpy as np
import pandas as pd

from argparse import SUPPRESS
from functools import partial
from shutil import copy
from itertools import izip
from pprint import pprint as ppr
from  multiprocessing import Pool,freeze_support

from telomerecat import core

######################################################################
##
##      Run the telomerecat length estimate on a CSV containing the
##      the following fields:
##          F1, F2, F4, Psi, Insert_mean, Insert_sd 
##
##      Author: jhrf
##
######################################################################

class LengthSimulator(object):
    def __init__(self,insert_mu,insert_sigma,
                    complete,boundary,
                    read_len=100):

        self._insert_mu = insert_mu
        self._insert_sigma = insert_sigma

        self._obs_total = complete + boundary #observed total
        self._read_thresh = int(self._obs_total * 0.0001)

        self._complete = complete 
        self._boundary = boundary
        self._read_len = read_len
        self._read_sim = self.__simulate_reads__

        self._get_read_len = (lambda: self._read_len)
        
    def start(self):
        if self._boundary <= 0 or self._complete <= 0:
            return 0
        else:
            return self.__run_sim__()

    def __run_sim__(self):
        found = False

        tel_len = int((float(self._complete)/self._boundary)\
                             * (self._insert_mu-100))
        max_its = 5000
        its = 0

        best_result = float("inf")
        best_tel_len  = tel_len

        read_simulator = self._read_sim
        get_factor = self.__get_factor__

        start = time.time()

        #print "TARGET: %.3f" % (self._obs_total * 0.0001,)

        while(not found and its < max_its):
            sim_comp,sim_boun,invalid_reads = read_simulator(tel_len)
            #   result can be positive or negative thus 
            #   influencing elongation and shortening
            #   A negative diference (overestimate) results 
            #   in a shorter telomere length next iteration
            result = self.is_same_ratio(sim_comp,sim_boun)
            abs_result = abs(result)

            if abs_result < best_result:
                best_result = abs_result
                best_tel_len = tel_len

            if result == 0: 
                found = True
            else:
                mult = -1 if result < 0 else 1
                factor = get_factor(abs_result)
                tel_len += factor * mult
            its += 1

            # if its % 10 == 0:
            #   print its,int(time.time()-start),"--------"
            #   print best_tel_len,best_result
            #   print tel_len,result

        return best_tel_len

    def __get_factor__(self,abs_result):
        total = self._obs_total

        if abs_result < (total * 0.001):
            factor = 1
        elif abs_result < (total * 0.2):
            factor = random.sample([1,2,2,4,25,200,500], 1)[0]
        elif abs_result < (total * 0.5):
            factor = random.sample([20,50,100,200,1000,2000], 1)[0]
        else:
            factor = int(np.log(abs_result)*300) / random.sample([1,2], 1)[0]
        return factor

    def is_same_ratio(self,comp,boun):
        difference = abs(comp - self._complete)
        if difference <= self._read_thresh:
            return 0
        else:
            return self._complete - comp

    def __simulate_reads__(self,tel_len):
        if tel_len < self._insert_mu:
            return int(self._complete * .50),self._boundary,0

        #speedup
        is_complete = self.is_complete
        is_boundary = self.is_boundary

        insert_mean = self._insert_mu
        insert_sigma = self._insert_sigma

        obs_total = self._obs_total
        #speedup

        est_complete = 0
        est_boundary = 0
        invalid_read = 0
        total = 0

        while total < (obs_total):
            insert_size = (random.gauss(insert_mean,insert_sigma))
            location = random.randint(0,tel_len)
            if is_complete(location,insert_size):
                est_complete += 1
                total += 1
            elif is_boundary(location,insert_size):
                est_boundary += 1
                total += 1
            else:
                invalid_read += 1

        return (est_complete,est_boundary,invalid_read)

    def is_complete(self,location,insert_size):
        return (location - insert_size) > 0

    def is_boundary(self,location,insert_size):
        return (location - self._get_read_len()) > 0

class Csv2Length(core.TelomerecatInterface):

    def __init__(self,
                 temp_dir=None,
                 total_procs=8,
                 verbose=False,
                 announce=True,
                 cmd_run=False):

        super(Csv2Length,self).__init__(instance_name = \
                                            "telomerecat csv2length", 
                                        temp_dir=temp_dir,
                                        total_procs=total_procs,
                                        verbose=verbose,
                                        announce = announce,
                                        cmd_run=cmd_run)

    def run_cmd(self):
        output_paths = self.__handle_cmd_outpaths__()

        self.run(input_paths = self.cmd_args.input,
                 correct_f2a = not self.cmd_args.disable_correction,
                 output_paths = output_paths,
                 prior_weight = self.cmd_args.prior_weight,
                 simulator_N = self.cmd_args.simulator_runs)

    def run(self,input_paths,
                 output_paths = [],
                 correct_f2a = True,
                 prior_weight = 3,
                 simulator_N = 10):
        
        self.__introduce__()
        self.__generate_output_paths__(input_paths, output_paths)

        self.__output__(" Commencing length estimation | %s\n" \
                            % (self.__get_date_time__(),),1)

        for input_path, output_path in izip(input_paths,output_paths):
            if self.announce:
                self.__output__("\tInput: %s\n" % (input_path,),1)
                self.__output__(" \tOutput: %s\n" % (output_path,),1)

            counts = pd.read_csv(input_path)
            counts = self.__get_length_from_dataframe__(counts,
                                                        correct_f2a,
                                                        prior_weight,
                                                        simulator_N)

            self.__output_length_results__(counts, output_path)
            self.__output__("\n",1)

        self.__goodbye__()
        return output_paths

    def __handle_cmd_outpaths__(self):
        output_paths = None
        if self.cmd_args.output is None:
            output_paths = []
        else:
            output_paths = self.cmd_args.output.split(",")
            output_paths = [p.strip() for p in output_paths]

        return output_paths

    def __get_length_from_dataframe__(self, counts,
                                        correct_f2a = True,
                                        prior_weight = 3,
                                        simulator_N = 10,
                                        simulate_lengths = True,):
        
        counts["F2a"]  = counts["F2"] - counts["F4"]

        if correct_f2a:
            counts["F2a_c"] = self.__get_corrected_F2a__(counts, prior_weight)
        else:
            counts["F2a_c"] = counts["F2a"]
        
        if simulate_lengths:
            counts["Length"] = self.__get_lengths__(counts, simulator_N)
        else:
            counts["Length"] = self.__quick_length__(counts)

        return counts

    def __get_corrected_F2a__(self, counts, prior_weight = 3):
        theta_observed = counts["F2a"]/ (counts["F2"] + counts["F4"])

        prior_weight = 3
        theta_expected = sum(theta_observed * counts["F2"]) / sum(counts["F2"])

        theta_corrected = ( (theta_observed * (counts["Psi"])) +\
                             (theta_expected * prior_weight)) \
                                / ((counts["Psi"]) + (prior_weight))

        corrected_f2_counts = (counts["F2"] + counts["F4"]) * theta_corrected

        return corrected_f2_counts.round(3)

    def __quick_length__(self, counts):
        lengths = []
        for i, sample in counts.iterrows():
            factor = sample["Read_length"] /\
                        (sample["F2a_c"]-sample["Read_length"])
            scale = sample["F2a_c"] + (sample["F2a_c"]*factor)
            length = ((sample["F1"] / scale) \
                        * sample["Insert_mean"]) + sample["Insert_mean"]
            lengths.append(round(length,3))
        return lengths

    def __get_lengths__(self,counts, simulator_N):
        lengths = []
        for i, sample in counts.iterrows():
            sample_intro = "\t- %s | %s\n" % (sample["Sample"], 
                                              self.__get_date_time__())
            self.__output__(sample_intro,2)

            length_mean,len_std = run_simulator_par(
                                        sample["Insert_mean"],
                                        sample["Insert_sd"],
                                        sample["F1"],
                                        sample["F2a_c"],
                                        self.total_procs,
                                        sample["Read_length"],
                                        simulator_N
                                        )
            lengths.append(length_mean)
        return lengths

    def __generate_output_paths__(self, input_paths, output_paths):
        if len(output_paths) < len(input_paths):
            for i in xrange(len(input_paths)):
                path = \
                    "./telomerecat_length_%d_%d.csv" % (time.time(), i)
                output_paths.append(path)

    def __output_length_results__(self, counts, count_path):
        counts.to_csv(count_path, index = None)

    def __copy_out_of_temp__(self,file_paths,copy_path="."):
        map(lambda fil: copy(fil,copy_path),file_paths)

    def get_parser(self):
        parser = self.default_parser()
        parser.description = textwrap.dedent(
        '''\
        %s
        %s

            The csv2length command allows the user to genereate new TL
            estimates from a previously generated CSV. This could be useful
            if the user wishes to re-run analysis with different options 
            (such as F2a correction) and does not wish to reanalyse
            the TELBAM files.

            Example useage:

            telomerecat csv2length /path/to/telomerecat_length_123.csv

            This will generate a new .csv file with revised telomere length
            estimates for the `telomerecat_length_123.csv` file.

        %s
        ''' % (self.instance_name, 
               self.header_line, 
               self.header_line,))

        parser.add_argument('input',metavar='CSV(s)', nargs='+',
            help="The CSV(s) that we wish to reanalyse")
        parser.add_argument('--output',
            metavar='CSV',type=str,nargs='?',default=None,
            help=('Specify output path for length estimation CSV.\n'
                  'If more than one input CSV is provided\n' 
                  'the user will need to provide a comma seperated list\n'
                  'with one entry per input. For example:\n\t'
                  '--output=\"zzz.csv, yyy.csv\". [Default: None]'))
        parser.add_argument('-s', help=SUPPRESS)
        parser.add_argument('-f', help=SUPPRESS)

        return parser

############################################
##
## These functions are implemented as global
## beacuse multiprocessing does not work with
## class bound methods
##
############################################

def check_results(sim_results):
    if 0 in sim_results:
        sys.stderr.write(
        "[WARNING] Telomere length reported zero. This means telomercat\n"+
        "\tfailed to identify enough complete or boundary reads.\n"+
        "\tThis  may mean your original sample was preprocessed to remove \n"+
        "\ttelomere reads. Alternatively this sample could have \n"+
        "\tvery short average TL.\n")

def run_simulator(insert_mu,insert_sigma,complete,boundary,proc,read_len,N=10):
    simmer = LengthSimulator(insert_mu,insert_sigma,complete,boundary,read_len)
    res = []
    invalid_count = []
    for i in range(N):
        length,invalid = simmer.start()
        res.append(length)
        invalid_count.append(invalid)

    check_results(res)
    return (np.mean(res),np.std(res))

def estimator_process(job,insert_mu,insert_sigma,complete,boundary,read_len):
    length_estimator = LengthSimulator(insert_mu,
                                       insert_sigma,
                                       complete,
                                       boundary,
                                       read_len)
    results = length_estimator.start()
    return results

def run_simulator_par(insert_mu,insert_sigma,complete,
                      boundary,proc,read_len,N=10):

    freeze_support()
    p = Pool(proc)
    sim_partial = partial(estimator_process,insert_mu=insert_mu,
                          insert_sigma=insert_sigma,complete=complete,
                          boundary=boundary,read_len=read_len)

    results = p.map(sim_partial,range(N))
    p.close()
    check_results(results)
    return (np.mean(results),np.std(results))

if __name__ == "__main__":
    print "Do not run this script directly. Type `telomerecat` for help."
